"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

# StatusCode must match pw::Status in pw_status.
class StatusCode(_StatusCode, metaclass=_StatusCodeEnumTypeWrapper):
    pass
class _StatusCode:
    V = typing.NewType('V', builtins.int)
class _StatusCodeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_StatusCode.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    OK = StatusCode.V(0)
    CANCELLED = StatusCode.V(1)
    UNKNOWN = StatusCode.V(2)
    INVALID_ARGUMENT = StatusCode.V(3)
    DEADLINE_EXCEEDED = StatusCode.V(4)
    NOT_FOUND = StatusCode.V(5)
    ALREADY_EXISTS = StatusCode.V(6)
    PERMISSION_DENIED = StatusCode.V(7)
    RESOURCE_EXHAUSTED = StatusCode.V(8)
    FAILED_PRECONDITION = StatusCode.V(9)
    ABORTED = StatusCode.V(10)
    OUT_OF_RANGE = StatusCode.V(11)
    UNIMPLEMENTED = StatusCode.V(12)
    INTERNAL = StatusCode.V(13)
    UNAVAILABLE = StatusCode.V(14)
    DATA_LOSS = StatusCode.V(15)
    UNAUTHENTICATED = StatusCode.V(16)

OK = StatusCode.V(0)
CANCELLED = StatusCode.V(1)
UNKNOWN = StatusCode.V(2)
INVALID_ARGUMENT = StatusCode.V(3)
DEADLINE_EXCEEDED = StatusCode.V(4)
NOT_FOUND = StatusCode.V(5)
ALREADY_EXISTS = StatusCode.V(6)
PERMISSION_DENIED = StatusCode.V(7)
RESOURCE_EXHAUSTED = StatusCode.V(8)
FAILED_PRECONDITION = StatusCode.V(9)
ABORTED = StatusCode.V(10)
OUT_OF_RANGE = StatusCode.V(11)
UNIMPLEMENTED = StatusCode.V(12)
INTERNAL = StatusCode.V(13)
UNAVAILABLE = StatusCode.V(14)
DATA_LOSS = StatusCode.V(15)
UNAUTHENTICATED = StatusCode.V(16)
global___StatusCode = StatusCode

