"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.descriptor_pb2
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.internal.extension_dict
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Tokenization(_Tokenization, metaclass=_TokenizationEnumTypeWrapper):
    pass
class _Tokenization:
    V = typing.NewType('V', builtins.int)
class _TokenizationEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Tokenization.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    # The field may contain plain text or any type of tokenized data (binary or
    # prefixed Base64).
    TOKENIZATION_OPTIONAL = Tokenization.V(0)

# The field may contain plain text or any type of tokenized data (binary or
# prefixed Base64).
TOKENIZATION_OPTIONAL = Tokenization.V(0)
global___Tokenization = Tokenization


# The field number was randomly selected from the reserved, internal use
# field numbers (50000-99999).
# TODO(pwbug/393): Register with the Protobuf Global Extension Registry:
#     https://github.com/protocolbuffers/protobuf/blob/HEAD/docs/options.md
format: google.protobuf.internal.extension_dict._ExtensionFieldDescriptor[google.protobuf.descriptor_pb2.FieldOptions, global___Tokenization.V] = ...
